#-----------------------
# @author: Tony Ribeiro
# @date: 2018/04/24
#
# Time series generator
# 
#-----------------------

from __future__ import print_function
import sys

import random
import math
from utils import eprint, generate_all_transitions, generate_given_transitions

#-----------------------
# rounding function
def roundup(x):
    return int(math.ceil(x / 10.0)) * 10
#-----------------------

# 0) Parameters management
#--------------------------

eprint(">> Starting System generation...")
eprint("> Parameter extraction...")

random.seed(a=None)

#******************
# parameters      *
#
nb_variables = "" 			# $1
nb_values = 2				# $2
semantic = "synchronous"	# $3
nb_transitions = 0			# $4
#                 *
#******************

if len(sys.argv) < 2:
	eprint("No number of variables specified")
	sys.exit()

nb_variables = int(sys.argv[1])
eprint("Number of variables given: ",nb_variables)
max_cond = nb_variables

if len(sys.argv) > 2:
	nb_values = int(sys.argv[2])
	eprint("Number of levels given: ",nb_values)
else:
	eprint("Default minimum number of levels: ",nb_values)
	
if len(sys.argv) > 3:
	semantic = sys.argv[3]
	if semantic != "synchronous" and semantic != "asynchronous" and semantic != "general":
		eprint("Parameter error: semantic must be one of synchronous, asynchronous or general")
		sys.exit()
	eprint("Semantic given: ",semantic)
else:
	eprint("Default semantic: "+semantic)
	
if len(sys.argv) > 4:
	nb_transitions = int(sys.argv[4])
	eprint("Number of transitions given: ",nb_transitions)
else:
	eprint("Default number of transitions, all: "+str(nb_values)+"^"+str(nb_variables)+"="+str(nb_values**nb_variables))

eprint ("Starting time series generator...")

# 1) Generation of the variables
#--------------------------------

eprint("> Generation of variables...")

variables = []
variables_values = []	

for i in range(0, nb_variables):
	variables.append("v"+str(i))
	variables_values.append([])
	for lvl in range(0,nb_values):
		variables_values[i].append(lvl)
		
# 2) Output variables
#------------------------------


for var in variables:
	print("VAR "+var,end='')
	for lvl in range(0,nb_values):
		print(" "+str(lvl),end='')
	print()
 
print()
print()
	
# 3) Generate random transitions
#--------------------------------

# Generate all transitions
if nb_transitions == 0:
	eprint("Generating all possible initial state")
	
	generate_all_transitions(variables, variables_values, 0, [], semantic)
	
# Generate given number of intitial states
else:
	eprint("Generating atleast "+str(nb_transitions)+" transitions")
	
	generate_given_transitions(variables, variables_values, 0, [], semantic, nb_transitions)
	
#############
# OUTPUT:
# - Variables/values
# - transitions
#############
