from __future__ import print_function
import sys
import random

#-----------------------
# Stderr print function
def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)
#-----------------------

#-----------------------
# Return a vector that gives for each variable the possible values concluded by the matching rules
def match(variables, state, rules):

	output = []

	for var in variables:
		output.append([])

	for rule in rules: # Check all rules
		#eprint("Check matching of ",rule.to_string(),"on",state)
		if rule.matches(state): # Matching conditions
			if rule.head_value not in output[rule.head_variable]: # New value
				output[rule.head_variable].append(rule.head_value)
	
	return output
#-----------------------

#-----------------------
# Generate a given number of transitions
def generate_random_transitions(variables, variables_values, semantic, nb_transitions):

	produced = 0
	while produced < nb_transitions:
	
		# Generate random initial state
		state = []
		for var in range(0, len(variables)):
			state.append(random.choice(variables_values[var]))
		
		# Generate random next state
		next_state = []
		for var in range(0, len(variables)):
			next_state.append(random.choice(variables_values[var]))
		
		transition_to_string(state, next_state, variables)
		print()
		produced+=1
		
	return


#-----------------------
# Generate all initial state and corresponding transitions
def generate_all_transitions(variables, variables_values, var, state, semantic):

	#eprint("Current state: ",state)

	# State fully initialized
	if var >= len(variables):
		#matching = match(variables,state,rules)
		matching = []
		# Generate a random matching
		for var_id in range(0,len(variables)):
			match=False
			while match == False: # Atleast one matching, complete program
				for val in variables_values[var_id]:
					if random.randint(0,1) == 1:
						matching.append((var_id,val))
						match=True
				
		
		#eprint("Random matching: "+str(matching))
		#sys.exit()
		
		# SYNCHRONOUS semantics: only one next state
		if semantic == "synchronous":
		
			# Ensure atleast one value per variable
			no_value = list(range(0,len(variables)))
			for i in matching:
				if i[0] in no_value:
					no_value.remove(i[0])
					
			#print("no value: "+str(no_value))
			for var_id in no_value:
				val = random.choice(variables_values[var_id])
				matching.append((var_id,val))
			#eprint("Matching: "+str(matching))
			
			combinations = subsets(matching)
			#eprint(combinations)
			
			next_states = []
			
			for combi in combinations:
				
				#print("checking combi: "+str(combi))
				
				# Check one value per variables
				if multiple_val_per_var(combi):
					continue
				
				# Check one value per variable
				if var_without_val(combi, variables):
					continue
		
				next_state = generate_transition(variables, state, combi)
				
				# Check duplicate
				if next_state in next_states:
					#eprint("duplicate: "+str(next_state))
					continue
					
				next_states.append(next_state)
					
				# Print new transition
				#eprint("new transition !")
				transition_to_string(state, next_state, variables)
				print()
				
			print()
		
			return
		
		# ASYNCHRONOUS semantics: only one variable can update in next state
		if semantic == "asynchronous":
		
			#eprint("Matching: "+str(matching))
		
			next_states = []
			possible_change = False
		
			# Generate one transition for each variable...
			for i in matching:
				var = i[0]
				val = i[1]
				#eprint("Change: "+str(var)+"="+str(val))
				
				# Check change
				if val == state[var]:
					continue
					
				possible_change = True
				
				next_state = list(state)
				next_state[var] = val
				
				# Check duplicate
				if next_state in next_states:
					#eprint("duplicate: "+str(next_state))
					continue
					
				next_states.append(next_state)
					
				# Print new transition
				#eprint("new transition !")
				transition_to_string(state, next_state, variables)
				print()
			
			# Point attractor when no change can occur
			if possible_change == False:
				transition_to_string(state, state, variables)
				print()
			
			print()
		
			return
			
		# GENERAL semantics: any number of variable can update in next state
		if semantic == "general":
		
			#eprint("Matching: "+str(matching))
		
			combinations = subsets(matching)
			#eprint(combinations)
			
			next_states = []
			
			for combi in combinations:
				
				# Check one value per variables
				if multiple_val_per_var(combi):
					continue
				
				next_state = generate_transition(variables, state, combi)
				
				# Check duplicate
				if next_state in next_states:
					#eprint("duplicate: "+str(next_state))
					continue
					
				next_states.append(next_state)
					
				# Print new transition
				#eprint("new transition !")
				transition_to_string(state, next_state, variables)
				print()
				
			print()
		
			return
		

	for val in variables_values[var]:
		state.append(val)
	
		# continue state initialization
		generate_all_transitions(variables, variables_values, var+1, state, semantic) 
		
		state.pop()
		
	return
#-----------------------

#-----------------------
# Generate all initial state and corresponding transitions until given limit
def generate_given_transitions(variables, variables_values, var, state, semantic, nb_transitions):
		
	#eprint("Current state: ",state)

	# State fully initialized
	if var >= len(variables):
		#matching = match(variables,state,rules)
		matching = []
		# Generate a random matching
		for var_id in range(0,len(variables)):
			match=False
			while match == False: # Atleast one matching, complete program
				for val in variables_values[var_id]:
					if random.randint(0,1) == 1:
						matching.append((var_id,val))
						match=True
				
		
		#eprint("Random matching: "+str(matching))
		#sys.exit()
		
		# SYNCHRONOUS semantics: only one next state
		if semantic == "synchronous":
		
			# Ensure atleast one value per variable
			no_value = list(range(0,len(variables)))
			for i in matching:
				if i[0] in no_value:
					no_value.remove(i[0])
					
			#print("no value: "+str(no_value))
			for var_id in no_value:
				val = random.choice(variables_values[var_id])
				matching.append((var_id,val))
			#eprint("Matching: "+str(matching))
			
			combinations = subsets(matching)
			#eprint(combinations)
			
			next_states = []
			
			for combi in combinations:
				
				#print("checking combi: "+str(combi))
				
				# Check one value per variables
				if multiple_val_per_var(combi):
					continue
				
				# Check one value per variable
				if var_without_val(combi, variables):
					continue
		
				next_state = generate_transition(variables, state, combi)
				
				# Check duplicate
				if next_state in next_states:
					#eprint("duplicate: "+str(next_state))
					continue
					
				next_states.append(next_state)
					
				# Print new transition
				#eprint("new transition !")
				transition_to_string(state, next_state, variables)
				print()
				nb_transitions-=1
				#eprint(nb_transitions)
			print()
		
			return nb_transitions
		
		# ASYNCHRONOUS semantics: only one variable can update in next state
		if semantic == "asynchronous":
		
			#eprint("Matching: "+str(matching))
		
			next_states = []
			possible_change = False
		
			# Generate one transition for each variable...
			for i in matching:
				var = i[0]
				val = i[1]
				#eprint("Change: "+str(var)+"="+str(val))
				
				# Check change
				if val == state[var]:
					continue
					
				possible_change = True
				
				next_state = list(state)
				next_state[var] = val
				
				# Check duplicate
				if next_state in next_states:
					#eprint("duplicate: "+str(next_state))
					continue
					
				next_states.append(next_state)
					
				# Print new transition
				#eprint("new transition !")
				transition_to_string(state, next_state, variables)
				print()
				nb_transitions-=1
				
			# Point attractor when no change can occur
			if possible_change == False:
				transition_to_string(state, state, variables)
				print()
				nb_transitions-=1
				#eprint(nb_transitions)
			print()
		
			return nb_transitions
			
		# GENERAL semantics: any number of variable can update in next state
		if semantic == "general":
		
			#eprint("Matching: "+str(matching))
		
			combinations = subsets(matching)
			#eprint(combinations)
			
			next_states = []
			
			for combi in combinations:
				
				# Check one value per variables
				if multiple_val_per_var(combi):
					continue
				
				next_state = generate_transition(variables, state, combi)
				
				# Check duplicate
				if next_state in next_states:
					#eprint("duplicate: "+str(next_state))
					continue
					
				next_states.append(next_state)
					
				# Print new transition
				#eprint("new transition !")
				transition_to_string(state, next_state, variables)
				print()
				nb_transitions-=1
				#eprint(nb_transitions)
			print()
		
			return nb_transitions
		

	for val in variables_values[var]:
		state.append(val)
	
		# continue state initialization
		nb_transitions = generate_given_transitions(variables, variables_values, var+1, state, semantic, nb_transitions)
		
		# All requested transitions done
		if nb_transitions <= 0:
			return nb_transitions
		
		state.pop()
		
	return nb_transitions
#-----------------------

def generate_transition(variables, state, combi):
	
	#eprint("Generating the next state for the combination:",combi)
	
	#if combi == []:
	#	return
	
	next_state = list(state)
	
	for i in combi:
		var = i[0]
		next_state[var] = i[1]
		
	return next_state

def subsets(s):
    sets = []
    for i in range(1 << len(s)):
        subset = [s[bit] for bit in range(len(s)) if is_bit_set(i, bit)]
        sets.append(subset)
    return sets

def is_bit_set(num, bit):
    return num & (1 << bit) > 0

def multiple_val_per_var(combi):

	double_value = False
	
	for i in range(0,len(combi)):
		for j in range(i+1,len(combi)):
			if combi[i][0] == combi[j][0]:
				double_value = True
				break
		if double_value:
			break
			
	return double_value
	
def var_without_val(combi, variables):

	no_value = range(0,len(variables))
	
	for i in combi:
		if i[0] in no_value:
			no_value.remove(i[0])
	
	#print("DBG: "+str(no_value))
	return len(no_value) > 0
	
def transition_to_string(state, next_state, variables):
	# Print current state
	for v in range(0,len(state)):
		print(variables[v]+"="+str(state[v])+" ", end='')
	print(": ", end='')
	
	# Print next state
	for v in range(0,len(next_state)):
		print(variables[v]+"="+str(next_state[v])+" ", end='')
